// Code Studio PoolMan Library

import com.codestudio.sql.PoolMan;
import com.codestudio.util.SQLResult;
import com.codestudio.util.SQLUtil;

import javax.sql.DataSource;
import java.sql.*;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * The sample methods in this file illustrate three possible uses
 * of PoolMan in JDBC work, listed in order of increasing sophistication:
 * <br>
 * SQLUtil: Use SQLUtil to execute SQL queries without using the JDBC API.
 * <br>
 * PoolMan Driver: Load the PoolMan JDBC Driver and use the JDBC API for
 * database access and operations.
 * <br>
 * PoolMan DataSource: Register PoolMan as a DataSource and subsequently
 * access it from a remote client, following the JDBC 2.0 standard.
 */
public class PoolManSample {


    public PoolManSample() {
    }


    /*
      ----------------
      SQLUTIL EXAMPLES
      ----------------
      The following method illustrates use of SQLUtil,
      which handles SQL queries quite easily but
      in a non-standard manner.

    */

    /**
     * Demo of the easiest way to use PoolMan -- use SQLUtil to get data
     * from a db without executing any JDBC (you still
     * have to know what SQL statement to execute, though, as
     * no object-relational mapping occurs).
     */
    public void doSQL(String sql, String dbname)
            throws SQLException {

        // init'ing once to avoid skewing the timers
        SQLResult preresult = SQLUtil.getInstance().execute(dbname, sql);

        long start = System.currentTimeMillis();

        SQLResult result = SQLUtil.getInstance().execute(dbname, sql);
        if (result.size() == 0)
            System.out.println("\tEmpty results");

        while (result.hasNext()) {
            Hashtable row = result.next();
            // in this "row", all keys are columns and all values are data
            for (Enumeration enum = row.keys(); enum.hasMoreElements();) {
                String column_name = enum.nextElement().toString();
                String data_field = row.get(column_name).toString();
                System.out.print("\t" + column_name + ": " + data_field);
            }
            System.out.print("\n");
        }

        System.out.println("\n\tExecution Time (millis): " +
                           (System.currentTimeMillis() - start));
    }

    /*
      ---------------------------
      POOLMAN JDBC DRIVER EXAMPLES
      ---------------------------
      The following three methods illustrate the use of PoolMan via
      standard JDBC. The approach simply calls upon the PoolMan driver
      to return Connections. The format of the URL supplied to the
      Driver must be "jdbc:poolman://[dbname]" where the dbname element
      corresponds to a dbname in your poolman.props file. You may omit
      the dbname parameter altogether, in which case the last database
      specified will be used.
    */

    /**
     * Demonstration of using standard JDBC and relying on the PoolMan
     * JDBC Driver to close all resources.
     */
    public void lazyQuery(String sql)
            throws SQLException {

        long start = System.currentTimeMillis();

        try {
            // load the PoolMan JDBC Driver
            Class.forName("com.codestudio.sql.PoolMan").newInstance();
        } catch (Exception ex) {
            System.out.println("Could Not Find the PoolMan Driver. " +
                               "Is poolman.jar in your CLASSPATH?");
            System.exit(0);
        }

        // establish a Connection to the first database listed in the 'poolman.xml' file
        Connection con = DriverManager.getConnection("jdbc:poolman");

        // just for the sake of example, create a scrollable ResultSet
        Statement s = con.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,
                                          ResultSet.CONCUR_READ_ONLY);

        ResultSet res = s.executeQuery(sql);
        ResultSetMetaData meta = res.getMetaData();

        int cols = meta.getColumnCount();

        // since it is scrollable, let's go backwards over the rows
        // start the cursor at the end
        res.afterLast();
        while (res.previous()) {
            for (int i = 1; i <= cols; i++) {
                Object val = res.getObject(i);
                System.out.print("\t" + meta.getColumnLabel(i) + ": ");
                System.out.print(val == null ? " " : val.toString());
            }
            System.out.print("\n");
        }

        // The Connection will return automatically to its pool and resources will be closed
        // so you the world won't end if you forget to do it explicitly here!

        System.out.println("\n\tExecution Time (millis): " +
                           (System.currentTimeMillis() - start));

    }

    /**
     * Demonstration of a cleaner JDBC standard API syntax,
     * explicitly closing all resources.
     */
    public void standardQuery(String sql)
            throws SQLException {

        long start = System.currentTimeMillis();

        // load the PoolMan JDBC Driver
        try {
            Class.forName("com.codestudio.sql.PoolMan").newInstance();
        } catch (Exception ex) {
            System.out.println("Could Not Find the PoolMan Driver. " +
                               "Is poolman.jar in your CLASSPATH?");
            System.exit(0);
        }

        // establish a Connection to the last database listed in the 'poolman.props' file
        Connection con = DriverManager.getConnection("jdbc:poolman");

        try {

            Statement s = con.createStatement();
            ResultSet res = s.executeQuery(sql);
            ResultSetMetaData meta = res.getMetaData();

            int cols = meta.getColumnCount();
            while (res.next()) {
                for (int i = 1; i <= cols; i++) {
                    Object val = res.getObject(i);
                    System.out.print("\t" + meta.getColumnLabel(i) + ": ");
                    System.out.print(val == null ? " " : val.toString());
                }
                System.out.print("\n");
            }
        } catch (SQLException sqe) {
        }
        finally {
            // this close method merely returns the Connection to the pool
            // after implicitly closing related resources (Statements and ResultSets)
            con.close();
            System.out.println("SAMPLE: Closed Con");
        }

        System.out.println("\n\tExecution Time (millis): " +
                           (System.currentTimeMillis() - start));
    }

    /**
     * Demonstration of a query in which PoolMan is using multiple databases.
     * This method uses traditional JDBC, but the same thing can be
     * accomplished by using the SQLUtil.execute(String dbname, String sql).
     */
    public void multipleDBQuery(String sql, String dbname)
            throws SQLException {

        long start = System.currentTimeMillis();

        // load the PoolMan JDBC Driver
        try {
            Class.forName("com.codestudio.sql.PoolMan").newInstance();
        } catch (Exception ex) {
            System.out.println("Could Not Find the PoolMan Driver. " +
                               "Is poolman.jar in your CLASSPATH?");
            System.exit(0);
        }

        // establish a Connection to the specified database
        Connection con = DriverManager.getConnection("jdbc:poolman://" + dbname);

        try {
            Statement s = con.createStatement();
            ResultSet res = s.executeQuery(sql);
            ResultSetMetaData meta = res.getMetaData();
            int cols = meta.getColumnCount();
            while (res.next()) {
                for (int i = 1; i <= cols; i++) {
                    Object val = res.getObject(i);
                    System.out.print("\t" + meta.getColumnLabel(i) + ": ");
                    System.out.print(val == null ? " " : val.toString());
                }
                System.out.print("\n");
            }
        } catch (SQLException sqe) {
        }
        finally {
            // this close method merely returns the Connection to the pool
            // and implicitly closes related Statements and ResultSets
            con.close();
        }

        System.out.println("\n\tExecution Time (millis): " +
                           (System.currentTimeMillis() - start));

    }

    /*
       --------------------------
       POOLMAN DATASOURCE EXAMPLE
       --------------------------
       The following method looks up PoolMan as a DataSource object.
     */

    public void doDataSourceExample(String dbName, String sql) {

        long start = System.currentTimeMillis();

        Connection con = null;

        try {

            DataSource ds = PoolMan.findDataSource(dbName);
            con = ds.getConnection();

            Statement st = con.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,
                                               ResultSet.CONCUR_READ_ONLY);

            ResultSet res = st.executeQuery(sql);

            ResultSetMetaData meta = res.getMetaData();
            int cols = meta.getColumnCount();
            // since this is a scrollable ResultSet,
            // do something a little strange...
            while (!res.isLast()) {
                res.next();
                for (int i = 1; i <= cols; i++) {
                    Object val = res.getObject(i);
                    System.out.print("\t" + meta.getColumnLabel(i) + ": ");
                    System.out.print(val == null ? " " : val.toString());
                }
                System.out.print("\n");
            }
        } catch (SQLException sqle) {
            sqle.printStackTrace();
        }
        finally {
            if (con != null)
                try {
                    con.close();
                } catch (SQLException sqle2) {
                }
        }

        System.out.println("\n\tExecution Time (millis): " +
                           (System.currentTimeMillis() - start));

    }


    public static void main(String args[]) {

        if (args.length < 2) {
            System.out.println("\nUSAGE NOTES: Include a sql string, in quotes, as a parameter " +
                               "and include a database name that matches a name " +
                               "specified in poolman.xml.");

            System.out.println("\nEXAMPLE: java PoolManSample \"select table_name from user_tables\" " +
                               "\"myDB\"");

            System.exit(0);
        }

        PoolManSample sample = new PoolManSample();

        try {

            System.out.println("\nExecuting sample of SQLUtil executing SQL from the specified database:\n");
            sample.doSQL(args[0], args[1]);
        } catch (Exception e) {
            e.printStackTrace();
        }


        try {
            System.out.println("\nExecuting sample of PoolMan JDBC executing SQL from the default database.");
            sample.standardQuery(args[0]);
        } catch (Exception e) {
            e.printStackTrace();
        }

        try {

            System.out.println("\nExecuting sample of PoolMan JDBC executing SQL from the specified database. " +
                               "The URL argument for this Connection will be \"jdbc:poolman://" + args[1] + "\"\n");
            sample.multipleDBQuery(args[0], args[1]);
        } catch (Exception e) {
            e.printStackTrace();
        }


        try {
            System.out.println("\nExecuting demo of PoolMan DataSource usage:\n");
            sample.doDataSourceExample(args[1], args[0]);

        } catch (Exception e) {
            e.printStackTrace();
        }

        System.out.println("\nPOOLMAN SAMPLES COMPLETE.");
        System.out.println("For more detailed examples, deploy the poolman-webapp in a servlet/JSP engine.");

        System.exit(0);

    }

}



